// ===================================+======================================
//
// Partitioner Client                                  Udo Sprute, 1995.10.27
//
// ==========================================================================

const char version [] = "Version 0.350";
const char program [] = "partnerd";
const char * jobname = "job";
unsigned niceness = 10;
unsigned delay = 1800;

# include <stdio.h>
# include <fcntl.h>
# include <signal.h>
# include <unistd.h>
# include <stdlib.h>
# include <string.h>
# include <iostream.h>
# include <sys/time.h>
# include <sys/stat.h>
# include <sys/wait.h>
# include <sys/types.h>
# include <sys/socket.h>

extern "C" {
   # include <netinet/in.h>
   # include <arpa/inet.h>
}

# undef FALSE
# undef TRUE
enum boolean {FALSE, TRUE};
boolean terminating = FALSE;
boolean testing = FALSE;

# ifndef SERVER
# define SERVER "127.0.0.1"
# endif

# ifndef PORT
# define PORT 4444
# endif

# ifndef PP
# define PP "g++ -Wall -O3"
# endif

# ifndef CC
# define CC "gcc -Wall -O3"
# endif

const char * server = SERVER;
unsigned short port = PORT;

# define DEBUG(x) (cerr << x << endl)

// --------------------------------------------------------------------------

pid_t child = 0;

void signalhandler (int signum)
{
   signal (signum, signalhandler);
   if (child)
      kill (child, SIGTERM);
   struct itimerval oldvalue;
   getitimer (ITIMER_REAL, & oldvalue);
   if (timerisset (& oldvalue.it_value)) {
      struct itimerval newvalue;
      newvalue.it_interval.tv_sec = 0;
      newvalue.it_interval.tv_usec = 0;
      newvalue.it_value.tv_sec = 0;
      newvalue.it_value.tv_usec = 1;
      setitimer (ITIMER_REAL, & newvalue, & oldvalue);
   }
   DEBUG ("Signal number " << signum << " caught.");
   DEBUG ("Terminating ...");
   terminating = TRUE;
}

void (* dummy0) (int) = signal (SIGPIPE, (void (*) (int)) SIG_IGN);
void (* dummy1) (int) = signal (SIGTERM, signalhandler);

void newhandler ()
{
   DEBUG ("No more memory.");
   exit (6);
}

extern "C" void (* set_new_handler (void (*) (void))) (void);
void (* dummy2) () = set_new_handler (newhandler);

// --------------------------------------------------------------------------

void sleepdelay (boolean reset)
{
   static unsigned seconds = 1;
   if (reset)
      seconds = 1;
   else {
      sleep (seconds);
      seconds = (seconds+seconds>delay) ? delay : seconds+seconds;
   }
}

// --------------------------------------------------------------------------

class string {
   char * value;
public:
   string (unsigned);
   string (const char * = NULL, const char * = NULL,
           const char * = NULL, const char * = NULL,
           const char * = NULL, const char * = NULL);
   string (unsigned, unsigned);
   operator char * ();
   char & operator [] (unsigned);
   char * operator + (unsigned);
   friend ostream & operator << (ostream &, string &);
   ~string ();
};

inline string::string (unsigned length)
{ value = new char [length]; }

string::string (
   const char * str0, const char * str1,
   const char * str2, const char * str3,
   const char * str4, const char * str5
) {
   unsigned length = 0;
   if (str0) { length += strlen (str0);
   if (str1) { length += strlen (str1);
   if (str2) { length += strlen (str2);
   if (str3) { length += strlen (str3);
   if (str4) { length += strlen (str4);
   if (str5) { length += strlen (str5);
   } } } } } }

   value = new char [length + 1];
   value [0] = '\0';
   if (str0) { strcat (value, str0);
   if (str1) { strcat (value, str1);
   if (str2) { strcat (value, str2);
   if (str3) { strcat (value, str3);
   if (str4) { strcat (value, str4);
   if (str5) { strcat (value, str5);
   } } } } } }
}

string::string (unsigned number, unsigned maximum)
{
   unsigned digits = 1;
   while (maximum /= 10)
      digits++;
   value = new char [digits + 1];
   value[digits] = '\0';
   while (digits--) {
      value[digits] = '0' + number%10;
      number /= 10;
   }
}

inline string::operator char * ()
{ return value; }

inline char & string::operator [] (unsigned n)
{ return value[n]; }

inline char * string::operator + (unsigned n)
{ return value + n; }

ostream & operator << (ostream & os, string & s)
{ return os << s.value; }

inline string::~string ()
{ delete [] value; }

// --------------------------------------------------------------------------

boolean mysystem (
   boolean searching,
   const char * command,
   const char * input = NULL,
   const char * output = NULL
) {
   child = fork ();
   if (child == -1)
      exit (5);

   if (child) {
      int status;
      if (wait(&status)==child && WIFEXITED(status) && !WEXITSTATUS(status)) {
         child = 0;
         return TRUE;
      } else {
         child = 0;
         return FALSE;
      }
   }

   if (input) {
      int fd = open (input, O_RDONLY);
      if (fd < 0) {
         cerr << "Cannot open " << '"' << input << '"' << " ";
         cerr << "for reading." << endl;
         exit (3);
      }
      if (fd != 0) {
         dup2 (fd, 0);
         close (fd);
      }
   }
   if (output) {
      int fd = creat (output, 0644);
      if (fd < 0) {
         cerr << "Cannot open " << '"' << output << '"' << " ";
         cerr << "for writing." << endl;
         exit (3);
      }
      if (fd != 1) {
         dup2 (fd, 1);
         close (fd);
      }
      dup2 (1, 2);
   }
   string coms (command);
   unsigned argc=0, n=0, i;
   for (i=0; coms[i]; ) {
      argc++;
      while (coms[i] && coms[i]!=' ' && coms[i]!='\t')
         i++;
      while (coms[i]==' ' || coms[i]=='\t')
         i++;
   }
   char ** argv = new char * [argc + 1];
   for (i=0; coms[i]; ) {
      argv[n++] = coms + i;
      while (coms[i] && coms[i]!=' ' && coms[i]!='\t')
         i++;
      while (coms[i]==' ' || coms[i]=='\t')
         coms[i++] = '\0';
   }
   argv[argc] = NULL;
   if (searching)
      execvp (argv[0], argv);
   else
      execv (argv[0], argv);
   delete [] argv;
   return FALSE;
}

// --------------------------------------------------------------------------

boolean test ()
{
   const char * program =
      "# include <stdio.h>\n"
      "int main () { printf (\"Hello.\\n\"); return 0; }\n";

   string sourcefile (jobname, ".c");
   int fd = creat (sourcefile, 0644);
   if (fd < 0)
      return FALSE;
   write (fd, program, strlen (program));
   close (fd);

   string compile (CC , " ", sourcefile, " -o ", jobname);
   return mysystem (TRUE, compile, "/dev/null");
}

// --------------------------------------------------------------------------

boolean daemon ()
{
   if (close (0) == 0) {
      string process (program, ".pid");
      struct stat fileinfo;
      if (stat (process, & fileinfo) == 0) {
         cerr << "Client process id file found: ";
         cerr << "another client seems to be running here." << endl;
         exit (9);
      }

      pid_t id = fork ();
      if (id == -1)
         exit (5);

      if (id) {
         int pfd = creat (process, 0644);
         if (pfd < 0) {
            cerr << "Cannot open " << '"' << process << '"' << " ";
            cerr << "for writing." << endl;
            exit (3);
         }
         unsigned n = 1;
         for (pid_t i=id; i/=10; )
            n++;
         string number (n+1);
         number[n] = '\0';
         while (n--) {
            number[n] = '0' + char (id%10);
            id /= 10;
         }
         write (pfd, number, strlen (number));
         write (pfd, "\n", 1);
         close (pfd);
         exit (0);
      }

      string file (program, ".log");
      int fd = open (file, O_CREAT | O_WRONLY | O_APPEND, 0644);
      if (fd < 0) {
         cerr << "Cannot open " << '"' << file << '"' << " ";
         cerr << "for writing." << endl;
         return FALSE;
      }
      for (unsigned n=0; n<256; n++)
         if (n != unsigned (fd))
            close (n);
      if (fd != 1) {
         dup2 (fd, 1);
         close (fd);
      }
      dup2 (1, 2);
      nice (niceness);
   }
   return TRUE;
}

// --------------------------------------------------------------------------

const unsigned BYTES = 4;

unsigned readnumber (int fd, boolean * ok)
{
   unsigned n = 0;
   if (sizeof n < BYTES) {
      cerr << "Decrease transmission length." << endl;
      exit (7);
   }
   if (read (fd, & n + sizeof n - BYTES, BYTES) == BYTES) {
      unsigned test = 1;
      if (((char *) &test) [0])
         for (unsigned i=0; i+i < sizeof n; i++) {
            ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
            ((char *) &n) [sizeof n - 1 - i] ^= ((char *) &n) [i];
            ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
         }
      * ok = TRUE;
   } else
      * ok = FALSE;
   return n;
}

boolean writenumber (int fd, unsigned n)
{
   if (sizeof n < BYTES) {
      cerr << "Decrease transmission length." << endl;
      exit (7);
   }
   unsigned test = 1;
   if (((char *) &test) [0])
      for (unsigned i=0; i+i < sizeof n; i++) {
         ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
         ((char *) &n) [sizeof n - 1 - i] ^= ((char *) &n) [i];
         ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
      }
   if (write (fd, & n + sizeof n - BYTES, BYTES) == BYTES)
      return TRUE;
   else
      return FALSE;
}

char * readline (int fd)
{
   boolean ok;
   size_t length = readnumber (fd, & ok);
   if (! ok)
      return NULL;
   char * buffer = new char [length+1];
   if (length && unsigned (read (fd, buffer, length)) != length)
      { delete [] buffer; return NULL; }
   buffer [length] = '\0';
   DEBUG ("---> " << '"' << buffer << '"');
   return buffer;
}

boolean writeline (int fd, const char * line)
{
   if (! line) line = "";
   size_t length = strlen (line);
   if (! writenumber (fd, length))
      return FALSE;
   if (length && unsigned (write (fd, line, length)) != length)
      return FALSE;
   DEBUG ("<--- " << '"' << line << '"');
   return TRUE;
}

boolean readfile (int fd, const char * file)
{
   boolean ok;
   size_t length = readnumber (fd, & ok);
   if (! ok)
      return FALSE;
   string buffer (length);
   unsigned transmitted = 0;
   while (transmitted < length) {
      int chunk = read (fd, buffer+transmitted, length-transmitted);
      if (chunk < 0)
         return FALSE;
      else
         transmitted += chunk;
   }

   if (length) {
      int ofd = creat (file, 0644);
      if (ofd < 0)
         return FALSE;
      if (unsigned (write (ofd, buffer, length)) != length)
         { close (ofd); unlink (file); return FALSE; }
      if (close (ofd) != 0)
         { unlink (file); return FALSE; }
   }

   DEBUG ("===> " << '"' << file << '"');
   return TRUE;
}

boolean writefile (int fd, const char * file)
{
   struct stat fileinfo;
   if (stat (file, & fileinfo) != 0)
      return FALSE;
   size_t length = size_t (fileinfo.st_size);

   string buffer (length);
   int ofd = open (file, O_RDONLY);
   if (ofd < 0)
      return FALSE;
   if (length && unsigned (read (ofd, buffer, length)) != length)
      { close (ofd); return FALSE; }
   if (close (ofd) != 0)
      return FALSE;

   if (! writenumber (fd, length))
      return FALSE;
   if (length && unsigned (write (fd, buffer, length)) != length)
      return FALSE;

   DEBUG ("<=== " << '"' << file << '"');
   return TRUE;
}

// --------------------------------------------------------------------------

# define COND if (good && !terminating)

boolean update (int fd)
{
   string oldprogram (program, ".old");
   boolean good = boolean (rename (program, oldprogram) == 0);

   string source (program, ".c");
   COND good = readfile (fd, source);

   string command (PP, " ", source, " -o ", program);
   COND good = mysystem (TRUE, command, "/dev/null");
   COND DEBUG ("xxxx " << '"' << command << '"');
      
   COND {
      string delayval (delay, delay);
      string delayopt ("-d", delayval);
      string jobopt ("-j", jobname);
      close (fd);
      unlink (oldprogram);
      DEBUG (":::: " <<
         '"' << program << " " << delayopt << " " << jobopt << '"'
      );
      DEBUG ("----------------");
      execl (program,
         program, (const char *) delayopt, (const char *) jobopt, NULL
      );
   } else {
      unlink (source);
      rename (oldprogram, program);
   }
   return FALSE;
}

void talk (unsigned long address, unsigned short port)
{
   static char * last = NULL;
   char * release = NULL;
   char * name = NULL;
   char * command = NULL;
   char * number = NULL;

   int fd = socket (AF_INET, SOCK_STREAM, 0);
   boolean good = boolean (fd >= 0);

   struct sockaddr_in target;
   target.sin_family = AF_INET;
   target.sin_port = htons (port);
   memcpy (& target.sin_addr, & address, sizeof target.sin_addr);

   COND good = boolean (
      connect (fd, (struct sockaddr *) & target, sizeof target) == 0
   );
   COND good = writeline (fd, version);
   COND good = boolean (release = readline (fd));
   COND if (*release && strcmp (version, release))
      good = update (fd);

   COND good = writeline (fd, last);
   COND good = boolean (name = readline (fd));
   COND good = boolean (* name);

   string sourcefile (jobname, ".c");
   string inputfile (jobname, ".in");
   string outputfile (jobname, ".out");
   COND if (!last || strcmp (last, name))
      good = readfile (fd, sourcefile);
   COND if (!last || strcmp (last, name))
      good = readfile (fd, inputfile);

   COND good = boolean (command = readline (fd));
   COND good = boolean (number = readline (fd));

   close (fd);

   COND if (!last || strcmp (last, name)) {
      string compile (CC , " ", sourcefile, " -o ", jobname);
      good = mysystem (TRUE, compile, "/dev/null");
      COND DEBUG ("xxxx " << '"' << compile << '"');
   }

   COND {
      string compute (jobname, " ", command);
      good = mysystem (FALSE, compute, inputfile, outputfile);
      COND DEBUG ("xxxx " << '"' << compute << '"');
      COND {
         delete [] last;
         last = new char [strlen (name) + 1];
         strcpy (last, name);
      } else {
         delete [] last;
         last = NULL;
      }
   }

   COND {
      struct stat fileinfo;
      if (stat (outputfile, & fileinfo) != 0)
         good = FALSE;
   }
   COND do {
      fd = socket (AF_INET, SOCK_STREAM, 0);
      good = boolean (fd >= 0);
      COND good = boolean (
         connect (fd, (struct sockaddr *) & target, sizeof target) == 0
      );
      COND good = writeline (fd, name);
      COND good = writeline (fd, number);
      COND good = writefile (fd, outputfile);
      close (fd);
      if (! terminating)
         sleepdelay (good);
   } while (! good && ! terminating);

   delete [] number;
   delete [] command;
   delete [] name;
   delete [] release;

   DEBUG ("----------------");
   if (! terminating)
      sleepdelay (good);
}

// --------------------------------------------------------------------------

int main (unsigned argc, char ** argv)
{
   unsigned n;
   for (n=1; n<argc && argv[n][0]=='-'; n++)
      switch (argv[n][1]) {
      case 'd':
         delay = atoi (argv[n]+2);
         break;
      case 'j':
         if (argv[n][2])
            jobname = argv[n] + 2;
         break;
      case 'n':
         niceness = atoi (argv[n]+2);
         break;
      case 't':
         testing = TRUE;
         break;
      case 'h':
      case '?':
         cout << "Usage: " << argv[0] << " ";
         cout << "[ -d<maxdelay> -j<jobname> -n<niceness> -t ] ";
         cout << "[ address [port] ]" << endl;
         cout << "Built-in defaults:" << endl;
         cout << "        SERVER = " << SERVER << endl;
         cout << "        PORT = " << PORT << endl;
         cout << "        PP = " << PP << endl;
         cout << "        CC = " << CC << endl;
         return 0;
      default:
         cerr << "Unknown option " << '"' << argv[n] << '"' << "." << endl;
         cerr << "Try " << '"' << argv[0] << " -?" << '"' << "." << endl;
         return 1;
      }
 
   switch (argc-n) {
   case 2:
      port = atoi (argv[n+1]);
   case 1:
      server = argv[n];
   case 0:
      break;
   default:
      cerr << "Too many parameters." << endl;
      cerr << "Try " << '"' << argv[0] << " -?" << '"' << "." << endl;
      return 1;
   }
 
   if (! getuid () || ! geteuid ()) {
      cerr << "I don't want to be run as root." << endl;
      cerr << "This is a security measure." << endl;
      return 2;
   }
   unsigned long address = inet_addr (server);
   if (address == unsigned (-1)) {
      cerr << "Invalid IP number address ";
      cerr << '"' << server << '"' << "." << endl;
      return 4;
   }
   if (testing && ! test ())
      return 8;
   if (! daemon ())
      return 3;
   while (! terminating)
      talk (address, port);

   string process (program, ".pid");
   unlink (process);
   return 0;
}

// ==========================================================================

