// ===================================+======================================
//
// Partitioner Server                                  Udo Sprute, 1995.10.27
//
// ==========================================================================

const char program [] = "partners";
const char clientc [] = "client.c";
const char * control = "jobs";
const unsigned nobackup = 19;

# include <stdio.h>
# include <fcntl.h>
# include <errno.h>
# include <signal.h>
# include <unistd.h>
# include <stdlib.h>
# include <string.h>
# include <iostream.h>
# include <sys/stat.h>
# include <sys/types.h>
# include <sys/socket.h>

extern "C" {
   # include <netinet/in.h>
}

# undef FALSE
# undef TRUE
enum boolean {FALSE, TRUE};
boolean terminating = FALSE;

# ifndef PORT
# define PORT 4444
# endif

unsigned short port = PORT;

# define DEBUG(x) (cerr << x << endl)

// --------------------------------------------------------------------------

void signalhandler (int signum)
{
   signal (signum, signalhandler);
   DEBUG ("Signal number " << signum << " caught.");
   DEBUG ("Terminating ...");
   terminating = TRUE;
}

void (* dummy0) (int) = signal (SIGPIPE, (void (*) (int)) SIG_IGN);
void (* dummy1) (int) = signal (SIGTERM, signalhandler);

void newhandler ()
{
   DEBUG ("No more memory.");
   exit (6);
}

extern "C" void (* set_new_handler (void (*) (void))) (void);
void (* dummy2) () = set_new_handler (newhandler);

// --------------------------------------------------------------------------

class string {
   char * value;
public:
   string (unsigned);
   string (const char * = NULL, const char * = NULL,
           const char * = NULL, const char * = NULL);
   operator char * ();
   char & operator [] (unsigned);
   char * operator + (unsigned);
   friend ostream & operator << (ostream &, string &);
   ~string ();
};

inline string::string (unsigned length)
{ value = new char [length]; }

string::string (
   const char * str0, const char * str1,
   const char * str2, const char * str3
) {
   unsigned length = 0;
   if (str0) { length += strlen (str0);
   if (str1) { length += strlen (str1);
   if (str2) { length += strlen (str2);
   if (str3) { length += strlen (str3);
   } } } }

   value = new char [length + 1];
   value [0] = '\0';
   if (str0) { strcat (value, str0);
   if (str1) { strcat (value, str1);
   if (str2) { strcat (value, str2);
   if (str3) { strcat (value, str3);
   } } } }
}

inline string::operator char * ()
{ return value; }

inline char & string::operator [] (unsigned n)
{ return value[n]; }

inline char * string::operator + (unsigned n)
{ return value + n; }

ostream & operator << (ostream & os, string & s)
{ return os << s.value; }

inline string::~string ()
{ delete [] value; }

// --------------------------------------------------------------------------

class clientsversion {
   char * name;
public:
   clientsversion ();
   void init (const char *);
   operator const char * ();
   ~clientsversion ();
};

clientsversion::clientsversion ()
{ name = NULL; }

void clientsversion::init (const char * file)
{
   struct stat fileinfo;
   if (stat (file, & fileinfo) != 0)
      return;
   unsigned length = unsigned (fileinfo.st_size);

   string buffer (length);
   int fd = open (file, O_RDONLY);
   if (fd < 0)
      return;
   if (unsigned (read (fd, buffer, length)) != length)
      { close (fd); return; }
   if (close (fd) != 0)
      return;

   unsigned m, n;
   for (m=0; m<length; m++)
      if (buffer[m] == '"')
         break;
   if (m < length)
      m++;
   for (n=m; n<length; n++)
      if (buffer[n] == '"')
         break;

   name = new char [n-m+1];
   for (unsigned i=0; i<n-m; i++)
      name[i] = buffer[m+i];
   name[n-m] = '\0';

   if (strstr (name, "Version") == name)
      cerr << "Partitioner " << '"' << name << '"' << endl;
   else {
      delete [] name;
      name = NULL;
   }
}

inline clientsversion::operator const char * ()
{ return name; }

inline clientsversion::~clientsversion ()
{ delete [] name; }

clientsversion version;

// --------------------------------------------------------------------------

class job {
   char * name;
   char * command;
   char * flags;
   unsigned numparts;
   unsigned numdigits;
   unsigned numnext;
   job * next;
   friend class joblist;
public:
   job (const char *, unsigned, unsigned, const char * = NULL, unsigned = 0);
   char * getjob ();
   void completiontest ();
   boolean setflag (unsigned, char);
   ~job ();
};

job::job (const char * line1, unsigned count1, unsigned parts,
          const char * line2, unsigned count2)
{
   unsigned n, i;

   for (n=0; n<count1; n++)
      if (line1[n] == ' ')
         break;

   name = new char [n+1];
   for (i=0; i<n; i++)
      name[i] = line1[i];
   name[n] = '\0';

   if (n<count1 && line1[n]==' ')
      n++;

   command = new char [count1-n+1];
   for (i=0; i<count1-n; i++)
      command[i] = line1[n+i];
   command[count1-n] = '\0';

   flags = new char [(count2<parts ? parts : count2) + 1];
   for (i=0; i<count2; i++)
      flags[i] = (line2[i]=='-') ? '.' : line2[i];
   while (i < parts)
      flags[i++] = '.';
   flags[i] = '\0';

   numparts = parts;

   if (numparts) {
      numdigits = 1;
      unsigned j = numparts - 1;
      while (j /= 10)
         numdigits++;
   } else
      numdigits = 0;

   numnext = 0;

   next = NULL;
}

char * job::getjob ()
{
   while (numnext<numparts && flags[numnext]!='.')
      numnext++;
   if (numnext == numparts)
      return NULL;

   unsigned digits = numdigits;
   unsigned number = numnext;
   char * result = new char [digits+1];
   result[digits] = '\0';
   while (digits--) {
      result[digits] = '0' + number%10;
      number /= 10;
   }
   return result;
}

boolean mogel = FALSE;

void job::completiontest ()
{
   for (char * cp=flags; *cp; cp++)
      if (*cp != '+')
         return;

   unsigned size = strlen ("output/") + strlen (name) + 1 + numdigits;
   string infile (size + 1);
   strcpy (infile, "output/");
   strcat (infile, name);
   strcat (infile, ".");
   for (unsigned i=numdigits; i; i--)
      infile [size-i] = '0';
   infile [size] = '\0';

   string file ("output/", name, ".out");
   int fd = open (file, O_CREAT | O_WRONLY | O_EXCL, 0644);
   if (fd >= 0)
      for (unsigned n=0; n<numparts; n++) {
         struct stat fileinfo;
         if (stat (infile, & fileinfo) == 0) {
            int fdin = open (infile, O_RDONLY);
            char buffer [1024];
            int count;
            while ((count = read (fdin, buffer, sizeof buffer)) > 0)
               write (fd, buffer, count);
            close (fdin);
            if (count == 0) unlink (infile);
         }
         unsigned j = size - 1;
         while (infile[j] == '9')
            infile[j--] = '0';
         infile[j] ++;
      }
   close (fd);
   
   delete [] flags;
   flags = new char [1];
   flags[0] = '\0';
   char * help = new char [strlen (command) + 2];
   strcpy (help, command);
   unsigned n = 0;
   while (command[n] && command[n]!='#')
      n++;
   if (command[n++] == '#') {
      help[n] = ' ';
      strcpy (help+n+1, command+n);
   }
   delete [] command;
   command = help;
   numnext = numparts = 0;
   mogel = TRUE;
}

boolean job::setflag (unsigned number, char c)
{
   if (number < numparts) {
      flags[number] = c;
      if (c=='+' && flags[numparts-1]=='+')
         completiontest ();
      return TRUE;
   } else {
      DEBUG ("Invalid access: " << '"' << name << '"' << " " << number);
      return FALSE;
   }
}

job::~job ()
{
   delete [] name;
   delete [] command;
   delete [] flags;
   delete next;
}

// --------------------------------------------------------------------------

class joblist {
   job * first;
   void insert (const char *, unsigned, unsigned, const char *, unsigned);
public:
   joblist ();
   void readcontrol (const char *);
   void writecontrol (const char *);
   void init (const char *);
   boolean known (const char *);
   boolean getjob (const char *, const char **, const char **, char **);
   boolean setflag (const char *, unsigned, char);
   ~joblist ();
};

void joblist::insert (const char * line1, unsigned count1, unsigned parts,
                      const char * line2, unsigned count2)
{
   unsigned n;
   for(n=0; n<count1; n++)
      if (line1[n] == ' ')
         break;
   string name (n+1);
   for (unsigned i=0; i<n; i++)
      name[i] = line1[i];
   name[n] = '\0';

   job ** current;
   for (current= &first; *current; current= &(*current)->next)
      if (! strcmp (name, (*current)->name)) {
         DEBUG ('"' << name << '"' << " project given twice, deleted once.");
         return;
      }
   * current = new job (line1, count1, parts, line2, count2);
}

void joblist::readcontrol (const char * file)
{
   struct stat fileinfo;
   if (stat (file, & fileinfo) != 0)
      return;
   unsigned length = unsigned (fileinfo.st_size);

   string buffer (length);
   int fd = open (file, O_RDONLY);
   if (fd < 0)
      return;
   if (unsigned (read (fd, buffer, length)) != length)
      { close (fd); return; }
   if (close (fd) != 0)
      return;

   for (unsigned n=0; n<length; n++) {
      unsigned i, j, k;
      for (i=0; n+i<length; i++)
         if (buffer[n+i] == '\n')
            break;
      if (! i)
         continue;

      boolean partsknown = FALSE;
      unsigned parts = 0;
      boolean valid = boolean (
         'A'<=buffer[n] && buffer[n]<='Z' ||
         'a'<=buffer[n] && buffer[n]<='z' ||
         buffer[n] == '_'
      );
      for (k=n+1; k<n+i; k++)
         if ('A'<=buffer[k] && buffer[k]<='Z' ||
             'a'<=buffer[k] && buffer[k]<='z' ||
             '0'<=buffer[k] && buffer[k]<='9' ||
             buffer[k] == '_')
            continue;
         else if (buffer[k]=='+' || buffer[k]=='-' || buffer[k]=='=')
            continue;
         else if (buffer[k] == ' ')
            continue;
         else if (buffer[k]=='#' && !partsknown) {
            parts = 0;
            while (k+1<n+i && '0'<=buffer[k+1] && buffer[k+1]<='9')
               parts = 10 * parts + buffer[++k] - '0';
            partsknown = TRUE;
         } else
            valid = FALSE;
      if (!partsknown || !valid) {
         n += i;
         continue;
      }

      for (j=0; (k=n+i+1+j)<length; j++)
         if (buffer[k]=='.' || buffer[k]=='-' || buffer[k]=='+')
            continue;
         else if (buffer[k] == '\n')
            break;
         else
            valid = FALSE;

      if (valid) {
         insert (buffer+n, i, parts, buffer+n+i+1, j);
         n += i+1+j;
      } else {
         insert (buffer+n, i, parts, NULL, 0);
         n += i;
      }
   }
}

void joblist::writecontrol (const char * file)
{
   int fd = creat (file, 0644);
   for (job * current=first; current; current=current->next) {
      write (fd, current->name, strlen (current->name));
      write (fd, " ", 1);
      write (fd, current->command, strlen (current->command));
      write (fd, "\n", 1);
      write (fd, current->flags, strlen (current->flags));
      write (fd, "\n", 1);
   }
   close (fd);
}

inline joblist::joblist ()
{ first = NULL; }

void joblist::init (const char * file)
{
   readcontrol (file);
   writecontrol (file);
}

boolean joblist::known (const char * name)
{
   for (job * current=first; current; current=current->next)
      if (! strcmp (name, current->name))
         return TRUE;
   return FALSE;
}

boolean joblist::getjob (
   const char * request,
   const char ** name,
   const char ** command,
   char ** number
) {
   job * current;
   for (current=first; current; current=current->next)
      if (! strcmp (request, current->name))
         if ((*number = current->getjob())) {
            *name = current->name;
            *command = current->command;
            return TRUE;
         }
   for (current=first; current; current=current->next)
      if ((*number = current->getjob())) {
         *name = current->name;
         *command = current->command;
         return TRUE;
      }
   return FALSE;
}

boolean joblist::setflag (const char * name, unsigned number, char c)
{
   for (job * current=first; current; current=current->next)
      if (! strcmp (name, current->name))
         if (current->setflag (number, c)) {
            static counter = 0;
            if (mogel || counter==nobackup) {
               writecontrol (control);
               mogel = FALSE;
               counter = 0;
            } else
               counter++;
            return TRUE;
         } else
            return FALSE;
   return FALSE;
}

inline joblist::~joblist ()
{ delete first; }

joblist jobs;

// --------------------------------------------------------------------------

boolean daemon (int sfd)
{
   string process (program, ".pid");
   struct stat fileinfo;
   if (stat (process, & fileinfo) == 0) {
      cerr << "Server process id file found: ";
      cerr << "another server seems to be running here." << endl;
      exit (9);
   }

   pid_t id = fork ();
   if (id == -1)
      exit (5);

   if (id) {
      int pfd = creat (process, 0644);
      if (pfd < 0) {
         cerr << "Cannot open " << '"' << process << '"' << " ";
         cerr << "for writing." << endl;
         exit (3);
      }
      unsigned n = 1;
      for (pid_t i=id; i/=10; )
         n++;
      string number (n+1);
      number[n] = '\0';
      while (n--) {
         number[n] = '0' + char (id%10);
         id /= 10;
      }
      write (pfd, number, strlen (number));
      write (pfd, "\n", 1);
      close (pfd);
      exit (0);
   }

   string file (program, ".log");
   int fd = open (file, O_CREAT | O_WRONLY | O_APPEND, 0644);
   if (fd < 0) {
      cerr << "Cannot open " << '"' << file << '"' << " ";
      cerr << "for writing." << endl;
      return FALSE;
   }
   for (unsigned n=0; n<256; n++)
      if (n!=unsigned(fd) && n!=unsigned(sfd))
         close (n);
   if (fd != 1) {
      dup2 (fd, 1);
      close (fd);
   }
   dup2 (1, 2);
   return TRUE;
}

// --------------------------------------------------------------------------

const unsigned BYTES = 4;

unsigned readnumber (int fd, boolean * ok)
{
   unsigned n = 0;
   if (sizeof n < BYTES) {
      cerr << "Decrease transmission length." << endl;
      exit (7);
   }
   if (read (fd, & n + sizeof n - BYTES, BYTES) == BYTES) {
      unsigned test = 1;
      if (((char *) &test) [0])
         for (unsigned i=0; i+i < sizeof n; i++) {
            ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
            ((char *) &n) [sizeof n - 1 - i] ^= ((char *) &n) [i];
            ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
         }
      * ok = TRUE;
   } else
      * ok = FALSE;
   return n;
}

boolean writenumber (int fd, unsigned n)
{
   if (sizeof n < BYTES) {
      cerr << "Decrease transmission length." << endl;
      exit (7);
   }
   unsigned test = 1;
   if (((char *) &test) [0])
      for (unsigned i=0; i+i < sizeof n; i++) {
         ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
         ((char *) &n) [sizeof n - 1 - i] ^= ((char *) &n) [i];
         ((char *) &n) [i] ^= ((char *) &n) [sizeof n - 1 - i];
      }
   if (write (fd, & n + sizeof n - BYTES, BYTES) == BYTES)
      return TRUE;
   else
      return FALSE;
}

char * readline (int fd)
{
   boolean ok;
   size_t length = readnumber (fd, & ok);
   if (! ok)
      return NULL;
   char * buffer = new char [length+1];
   if (length && unsigned (read (fd, buffer, length)) != length)
      { delete [] buffer; return NULL; }
   buffer [length] = '\0';
   DEBUG ("---> " << '"' << buffer << '"');
   return buffer;
}

boolean writeline (int fd, const char * line)
{
   if (! line) line = "";
   size_t length = strlen (line);
   if (! writenumber (fd, length))
      return FALSE;
   if (length && unsigned (write (fd, line, length)) != length)
      return FALSE;
   DEBUG ("<--- " << '"' << line << '"');
   return TRUE;
}

boolean readfile (int fd, const char * file)
{
   boolean ok;
   size_t length = readnumber (fd, & ok);
   if (! ok)
      return FALSE;
   string buffer (length);
   unsigned transmitted = 0;
   while (transmitted < length) {
      int chunk = read (fd, buffer+transmitted, length-transmitted);
      if (chunk < 0)
         return FALSE;
      else
         transmitted += chunk;
   }

   if (length) {
      int ofd = creat (file, 0644);
      if (ofd < 0)
         return FALSE;
      if (unsigned (write (ofd, buffer, length)) != length)
         { close (ofd); unlink (file); return FALSE; }
      if (close (ofd) != 0)
         { unlink (file); return FALSE; }
   }

   DEBUG ("===> " << '"' << file << '"');
   return TRUE;
}

boolean writefile (int fd, const char * file)
{
   struct stat fileinfo;
   if (stat (file, & fileinfo) != 0)
      return FALSE;
   size_t length = size_t (fileinfo.st_size);

   string buffer (length);
   int ofd = open (file, O_RDONLY);
   if (ofd < 0)
      return FALSE;
   if (length && unsigned (read (ofd, buffer, length)) != length)
      { close (ofd); return FALSE; }
   if (close (ofd) != 0)
      return FALSE;

   if (! writenumber (fd, length))
      return FALSE;
   if (length && unsigned (write (fd, buffer, length)) != length)
      return FALSE;

   DEBUG ("<=== " << '"' << file << '"');
   return TRUE;
}

// --------------------------------------------------------------------------

# define COND if (good && !terminating)

void talk (int socketfd)
{
   char * release = NULL;
   char * request = NULL;
   char * number = NULL;

   struct sockaddr sourceaddress;
   int sourceaddresslength = sizeof sourceaddress;
   int fd = accept (socketfd, & sourceaddress, & sourceaddresslength);
   boolean good = boolean (fd >= 0);
   COND DEBUG ("#### " <<
      unsigned ((unsigned char) sourceaddress.sa_data[2]) << "." <<
      unsigned ((unsigned char) sourceaddress.sa_data[3]) << "." <<
      unsigned ((unsigned char) sourceaddress.sa_data[4]) << "." <<
      unsigned ((unsigned char) sourceaddress.sa_data[5])
   );

   COND good = boolean (release = readline (fd));
   COND if (strstr (release, "Version") == release) {
      COND good = writeline (fd, version);
      COND if ((const char *)(version) && strcmp (version, release)) {
         writefile (fd, clientc);
         good = FALSE;
      }
      COND good = boolean (request = readline (fd));
      const char * name = NULL;
      const char * command = NULL;
      COND {
         good = jobs.getjob (request, & name, & command, & number);
         if (! writeline (fd, name))
            good = FALSE;
      }
      COND if (strcmp (request, name)) {
         string sourcefile ("input/", name, ".c");
         good = writefile (fd, sourcefile);
      }
      COND if (strcmp (request, name)) {
         string inputfile ("input/", name, ".in");
         good = writefile (fd, inputfile);
      }
      COND {
         string todo (strlen (command) + 1);
         strcpy (todo, command);
         unsigned i = 0;
         while (todo[i] != '#')
            i++;
         todo [i++] = '\0';
         while (todo[i] && '0'<=todo[i] && todo[i]<='9')
            i++;
         strcat (todo, number); 
         strcat (todo, command+i);
         good = writeline (fd, todo);
      }
      COND good = writeline (fd, number);
      COND good = jobs.setflag (name, atoi (number), '-');
   } else {
      COND request = release, release = NULL;
      COND good = jobs.known (request);
      COND good = boolean (number = readline (fd));
      COND for (const char * cp=number; *cp; cp++)
         if (*cp<'0' || '9'<*cp)
            good = FALSE;
      string outputfile ("output/", request, ".", number);
      COND good = readfile (fd, outputfile);
      COND good = jobs.setflag (request, atoi (number), '+');
   }

   close (fd);
   delete [] number;
   delete [] request;
   delete [] release;
   DEBUG ("----------------");
}

// --------------------------------------------------------------------------

const unsigned LISTENQUEUE = 5;

int main (unsigned argc, char ** argv)
{
   unsigned n;
   for (n=1; n<argc && argv[n][0]=='-'; n++)
      switch (argv[n][1]) {
      case 'f':
         if (argv[n][2])
            control = argv[n] + 2;
         break;
      case 'h':
      case '?':
         cout << "Usage: " << argv[0] << " ";
         cout << "[ -f<filename> ] ";
         cout << "[ port ]" << endl;
         cout << "Built-in defaults:" << endl;
         cout << "        PORT = " << PORT << endl;
         return 0;
      default:
         cerr << "Unknown option " << '"' << argv[n] << '"' << "." << endl;
         cerr << "Try " << '"' << argv[0] << " -?" << '"' << "." << endl;
         return 1;
      }
 
   switch (argc-n) {
   case 1:
      port = atoi (argv[n]);
   case 0:
      break;
   default:
      cerr << "Too many parameters." << endl;
      cerr << "Try " << '"' << argv[0] << " -?" << '"' << "." << endl;
      return 1;
   }

   version.init (clientc);
   jobs.init (control);

   if (! getuid () || ! geteuid ()) {
      cerr << "I don't want to be run as root." << endl;
      cerr << "This is a security measure." << endl;
      return 2;
   }
   if (mkdir ("input", 0755), chdir ("input") || chdir ("..")) {
      cerr << "Cannot create subdirectory ";
      cerr << '"' << "input" << '"' << "." << endl;
      return 3;
   }
   if (mkdir ("output", 0755), chdir ("output") || chdir ("..")) {
      cerr << "Cannot create subdirectory ";
      cerr << '"' << "output" << '"' << "." << endl;
      return 3;
   }

   struct sockaddr_in target;
   target.sin_family = AF_INET;
   target.sin_port = htons (port);
   unsigned long address = 0;
   memcpy (& target.sin_addr, & address, sizeof target.sin_addr);

   int fd = socket (AF_INET, SOCK_STREAM, 0);
   if (fd < 0) {
      perror ("Socket");
      return 4;
   }
   if (bind (fd, (struct sockaddr *) & target, sizeof target) != 0) {
      perror ("Bind");
      return 4;
   }
   if (listen (fd, LISTENQUEUE) != 0) {
      perror ("Listen");
      return 4;
   }
   if (! daemon (fd))
      return 3;
   while (! terminating)
      talk (fd);
   close (fd);

   jobs.writecontrol (control);
   string process (program, ".pid");
   unlink (process);
   return 0;
}

// ==========================================================================

